import os
from ctypes import Structure, POINTER, c_int, c_float, c_void_p, c_char, \
    c_char_p, c_double
from ctypes import POINTER as _P
from .dll import DLL, SDLFunc
from .stdinc import Uint8, Uint32, Sint16
from .render import SDL_Renderer
from .surface import SDL_Surface

# NOTE: This module is currently missing wrappers for the image filtering
# functions in SDL2_imageFilter.h. However, because we have Pillow on Python
# this isn't really a pressing concern. Time permitting, these functions may
# be wrapped at a later date for the sake of completeness.

__all__ = [
    # Structs
    "FPSManager",
    
    # Defines
    "FPS_UPPER_LIMIT", "FPS_LOWER_LIMIT", "FPS_DEFAULT",
    "SDL2_GFXPRIMITIVES_MAJOR", "SDL2_GFXPRIMITIVES_MAJOR",
    "SDL2_GFXPRIMITIVES_MICRO", "SMOOTHING_OFF", "SMOOTHING_ON",

    # Functions
    "SDL_initFramerate", "SDL_getFramerate",
    "SDL_setFramerate", "SDL_getFramecount", "SDL_framerateDelay",
    "pixelColor", "pixelRGBA", "hlineColor",
    "hlineRGBA", "vlineColor", "vlineRGBA", "rectangleColor",
    "rectangleRGBA", "roundedRectangleColor", "roundedRectangleRGBA",
    "boxColor", "boxRGBA", "roundedBoxColor", "roundedBoxRGBA",
    "lineColor", "lineRGBA", "aalineColor", "aalineRGBA",
    "thickLineColor", "thickLineRGBA", "circleColor", "circleRGBA",
    "arcColor", "arcRGBA", "aacircleColor", "aacircleRGBA",
    "filledCircleColor", "filledCircleRGBA", "ellipseColor",
    "ellipseRGBA", "aaellipseColor", "aaellipseRGBA",
    "filledEllipseColor", "filledEllipseRGBA", "pieColor", "pieRGBA",
    "filledPieColor", "filledPieRGBA", "trigonColor", "trigonRGBA",
    "aatrigonColor", "aatrigonRGBA", "filledTrigonColor",
    "filledTrigonRGBA", "polygonColor", "polygonRGBA", "aapolygonColor",
    "aapolygonRGBA", "filledPolygonColor", "filledPolygonRGBA",
    "texturedPolygon", "bezierColor", "bezierRGBA",
    "gfxPrimitivesSetFont", "gfxPrimitivesSetFontRotation",
    "characterColor", "characterRGBA", "stringColor", "stringRGBA",
    "rotozoomSurface", "rotozoomSurfaceXY", "rotozoomSurfaceSize",
    "rotozoomSurfaceSizeXY", "zoomSurface", "zoomSurfaceSize", "shrinkSurface",
    "rotateSurface90Degrees",

    # Python Functions
    "get_dll_file"
]


try:
    dll = DLL("SDL2_gfx", ["SDL2_gfx", "SDL2_gfx-1.0"],
              os.getenv("PYSDL2_DLL_PATH"))
except RuntimeError as exc:
    raise ImportError(exc)


def get_dll_file():
    """Gets the file name of the loaded SDL2_gfx library."""
    return dll.libfile

_bind = dll.bind_function


# Constants, enums, type definitions, and macros

SDL2_GFXPRIMITIVES_MAJOR = 1
SDL2_GFXPRIMITIVES_MINOR = 0
SDL2_GFXPRIMITIVES_MICRO = 4

FPS_UPPER_LIMIT = 200
FPS_LOWER_LIMIT = 1
FPS_DEFAULT = 30

SMOOTHING_OFF = 0
SMOOTHING_ON = 1

class FPSManager(Structure):
    """A structure holding the state and timing of the framerate manager.

    This class can be used with other SDL_gfx functions to set a custom
    framerate within a given rendering loop. When used with
    :func:`SDL_framerateDelay`, it uses its initial frame onset time
    (:attr:`baseticks`) and the duration per frame to try to present frames
    at consistent intervals from that initial point.

    .. note::
        This method of frame pacing may not play nicely with vsync in SDL2.

    Attributes:
        framecount (int): The number of frames counted by the manager since
            being initialized.
        rateticks (float): The time delay (in ms) between each frame.
        baseticks (int): The milliseconds since SDL initialization at which the
            manager was initialized with :func:`SDL_initFramerate`. Used
            internally as the initial frame onset time.
        lastticks (int): The milliseconds since SDL initialization at which the
            previous frame was displayed.
        rate (int): The framerate (in Hz) of the manager.

    """
    _fields_ = [("framecount", Uint32),
                ("rateticks", c_float),
                ("baseticks", Uint32),
                ("lastticks", Uint32),
                ("rate", Uint32)
                ]


# Raw ctypes function definitions

_funcdefs = [
    SDLFunc("SDL_initFramerate", [_P(FPSManager)]),
    SDLFunc("SDL_setFramerate", [_P(FPSManager), Uint32], c_int),
    SDLFunc("SDL_getFramerate", [_P(FPSManager)], c_int),
    SDLFunc("SDL_getFramecount", [_P(FPSManager)], Uint32),
    SDLFunc("SDL_framerateDelay", [_P(FPSManager)], Uint32),
    SDLFunc("pixelColor", [_P(SDL_Renderer), Sint16, Sint16, Uint32], c_int),
    SDLFunc("pixelRGBA", [_P(SDL_Renderer), Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("hlineColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("hlineRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("vlineColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("vlineRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("rectangleColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("rectangleRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("roundedRectangleColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("roundedRectangleRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("boxColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("boxRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("roundedBoxColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("roundedBoxRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("lineColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("lineRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("aalineColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("aalineRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("thickLineColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint8, Uint32], c_int),
    SDLFunc("thickLineRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("circleColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("circleRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("arcColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("arcRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("aacircleColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("aacircleRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("filledCircleColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("filledCircleRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("ellipseColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("ellipseRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("aaellipseColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("aaellipseRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("filledEllipseColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("filledEllipseRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("pieColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("pieRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("filledPieColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("filledPieRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("trigonColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("trigonRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("aatrigonColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("aatrigonRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("filledTrigonColor", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Sint16, Uint32], c_int),
    SDLFunc("filledTrigonRGBA", [_P(SDL_Renderer), Sint16, Sint16, Sint16, Sint16, Sint16, Sint16, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("polygonColor", [_P(SDL_Renderer), _P(Sint16), _P(Sint16), c_int, Uint32], c_int),
    SDLFunc("polygonRGBA", [_P(SDL_Renderer), _P(Sint16), _P(Sint16), c_int, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("aapolygonColor", [_P(SDL_Renderer), _P(Sint16), _P(Sint16), c_int, Uint32], c_int),
    SDLFunc("aapolygonRGBA", [_P(SDL_Renderer), _P(Sint16), _P(Sint16), c_int, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("filledPolygonColor", [_P(SDL_Renderer), _P(Sint16), _P(Sint16), c_int, Uint32], c_int),
    SDLFunc("filledPolygonRGBA", [_P(SDL_Renderer), _P(Sint16), _P(Sint16), c_int, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("texturedPolygon", [_P(SDL_Renderer), _P(Sint16), _P(Sint16), c_int, _P(SDL_Surface), c_int, c_int], c_int),
    SDLFunc("bezierColor", [_P(SDL_Renderer), _P(Sint16), _P(Sint16), c_int, c_int, Uint32], c_int),
    SDLFunc("bezierRGBA", [_P(SDL_Renderer), _P(Sint16), _P(Sint16), c_int, c_int, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("gfxPrimitivesSetFont", [c_void_p, Uint32, Uint32]),
    SDLFunc("gfxPrimitivesSetFontRotation", [Uint32]),
    SDLFunc("characterColor", [_P(SDL_Renderer), Sint16, Sint16, c_char, Uint32], c_int),
    SDLFunc("characterRGBA", [_P(SDL_Renderer), Sint16, Sint16, c_char, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("stringColor", [_P(SDL_Renderer), Sint16, Sint16, c_char_p, Uint32], c_int),
    SDLFunc("stringRGBA", [_P(SDL_Renderer), Sint16, Sint16, c_char_p, Uint8, Uint8, Uint8, Uint8], c_int),
    SDLFunc("rotozoomSurface", [_P(SDL_Surface), c_double, c_double, c_int], _P(SDL_Surface)),
    SDLFunc("rotozoomSurfaceXY", [_P(SDL_Surface), c_double, c_double, c_double, c_int], _P(SDL_Surface)),
    SDLFunc("rotozoomSurfaceSize", [c_int, c_int, c_double, c_double, _P(c_int), _P(c_int)]),
    SDLFunc("rotozoomSurfaceSizeXY", [c_int, c_int, c_double, c_double, c_double, _P(c_int), _P(c_int)]),
    SDLFunc("zoomSurface", [_P(SDL_Surface), c_double, c_double, c_int], _P(SDL_Surface)),
    SDLFunc("zoomSurfaceSize", [c_int, c_int, c_double, c_double, _P(c_int), _P(c_int)]),
    SDLFunc("shrinkSurface", [_P(SDL_Surface), c_int, c_int], _P(SDL_Surface)),
    SDLFunc("rotateSurface90Degrees", [_P(SDL_Surface), c_int], _P(SDL_Surface)),
]
_funcs = {}
for f in _funcdefs:
    _funcs[f.name] = _bind(f.name, f.args, f.returns, f.added)


# Python wrapper functions

def SDL_initFramerate(manager):
    """Initializes a framerate manager.

    Calling this function on an :class:`FPSManager` initializes it with a
    default framerate of 30 Hz and prepares it for counting and timing frames.
    If the manager was already initialized, calling this function will reset
    its framecount, initial frame onset time, and framerate.

    Args:
        manager (:obj:`sdlgfx.FPSManager`): The framerate manager to initialize.

    """
    return _funcs["SDL_initFramerate"](manager)

def SDL_setFramerate(manager, rate):
    """Sets the framerate of a framerate manager.

    Sets a new framerate for the manager, resetting both the framecount and the
    the initial frame onset time. Framerates must be between ``FPS_LOWER_LIMIT``
    (1) and ``FPS_UPPER_LIMIT`` (200), inclusive, to be accepted.

    Args:
        manager (:obj:`sdlgfx.FPSManager`): The framerate manager to configure.
        rate (int): The new framerate in Hz.

    Returns:
        int: 0 on success, or -1 if an error occurred.

    """
    return _funcs["SDL_setFramerate"](manager, rate)

def SDL_getFramerate(manager):
    """Gets the current framerate for a framerate manager.

    Args:
        manager (:obj:`sdlgfx.FPSManager`): The framerate manager for which the
            currently set framerate will be retrieved.

    Returns:
        int: 0 on success, or -1 if an error occurred.

    """
    return _funcs["SDL_getFramerate"](manager)

def SDL_getFramecount(manager):
    """Gets the current number of frames counted by a framerate manager.

    .. note::
        This value is reset whenever a frame is dropped (i.e. the rendering
        loop takes longer than the set interval between frames) or the framerate
        is changed.

    Args:
        manager (:obj:`sdlgfx.FPSManager`): The framerate manager for which the
            current framecount will be retrieved.

    Returns:
        int: 0 on success, or -1 if an error occurred.

    """
    return _funcs["SDL_getFramecount"](manager)

def SDL_framerateDelay(manager):
    """Delays execution until the next frame occurs.

    This function waits for the next frame onset (as determined by the rate set
    by :func:`SDL_setFramerate`) to keep frame pacing consistent. This should be
    called once per loop within the program's main rendering loop.
    
    If the rendering loop takes longer than the set framerate, the delay will be
    zero and the framecount and initial frame onset time will be reset.

    Args:
        manager (:obj:`sdlgfx.FPSManager`): The framerate manager to use for
            frame pacing.

    Returns:
        int: 0 on success, or -1 if an error occurred.

    """
    return _funcs["SDL_framerateDelay"](manager)


def pixelColor(renderer, x, y, color):
    """Draws a single pixel to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X (horizontal) coordinate of the pixel.
        y (int): The Y (vertical) coordinate of the pixel.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["pixelColor"](renderer, x, y, color)

def pixelRGBA(renderer, x, y, r, g, b, a):
    """Draws a single pixel to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X (horizontal) coordinate of the pixel.
        y (int): The Y (vertical) coordinate of the pixel.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["pixelRGBA"](renderer, x, y, r, g, b, a)

def hlineColor(renderer, x1, x2, y, color):
    """Draws a horizontal line to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the first point of the line.
        x2 (int): The X coordinate of the second point of the line.
        y (int): The Y (vertical) coordinate of the points of the line.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["hlineColor"](renderer, x1, x2, y, color)

def hlineRGBA(renderer, x1, x2, y, r, g, b, a):
    """Draws a horizontal line to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the first point of the line.
        x2 (int): The X coordinate of the second point of the line.
        y (int): The Y coordinate of the points of the line.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["hlineRGBA"](renderer, x1, x2, y, r, g, b, a)

def vlineColor(renderer, x, y1, y2, color):
    """Draws a vertical line to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the points of the line.
        y1 (int): The X coordinate of the first point of the line.
        y2 (int): The Y coordinate of the second point of the line.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["vlineColor"](renderer, x, y1, y2, color)

def vlineRGBA(renderer, x, y1, y2, r, g, b, a):
    """Draws a vertical line to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the points of the line.
        y1 (int): The X coordinate of the first point of the line.
        y2 (int): The Y coordinate of the second point of the line.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["vlineRGBA"](renderer, x, y1, y2, r, g, b, a)


def rectangleColor(renderer, x1, y1, x2, y2, color):
    """Draws an unfilled rectangle to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the top-right point of the rectangle.
        y1 (int): The Y coordinate of the top-right point of the rectangle.
        x2 (int): The X coordinate of the bottom-left point of the rectangle.
        y2 (int): The Y coordinate of the bottom-left point of the rectangle.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["rectangleColor"](renderer, x1, y1, x2, y2, color)

def rectangleRGBA(renderer, x1, y1, x2, y2, r, g, b, a):
    """Draws an unfilled rectangle to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the top-right point of the rectangle.
        y1 (int): The Y coordinate of the top-right point of the rectangle.
        x2 (int): The X coordinate of the bottom-left point of the rectangle.
        y2 (int): The Y coordinate of the bottom-left point of the rectangle.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["rectangleRGBA"](renderer, x1, y1, x2, y2, r, g, b, a)

def roundedRectangleColor(renderer, x1, y1, x2, y2, rad, color):
    """Draws an unfilled rectangle with rounded corners to the renderer.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the top-right point of the rectangle.
        y1 (int): The Y coordinate of the top-right point of the rectangle.
        x2 (int): The X coordinate of the bottom-left point of the rectangle.
        y2 (int): The Y coordinate of the bottom-left point of the rectangle.
        rad (int): The radius of the arc of the rounded corners.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["roundedRectangleColor"](renderer, x1, y1, x2, y2, rad, color)

def roundedRectangleRGBA(renderer, x1, y1, x2, y2, rad, r, g, b, a):
    """Draws an unfilled rectangle with rounded corners to the renderer.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the top-right point of the rectangle.
        y1 (int): The Y coordinate of the top-right point of the rectangle.
        x2 (int): The X coordinate of the bottom-left point of the rectangle.
        y2 (int): The Y coordinate of the bottom-left point of the rectangle.
        rad (int): The radius of the arc of the rounded corners.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["roundedRectangleRGBA"](renderer, x1, y1, x2, y2, rad, r, g, b, a)

def boxColor(renderer, x1, y1, x2, y2, color):
    """Draws a filled rectangle to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the top-right point of the rectangle.
        y1 (int): The Y coordinate of the top-right point of the rectangle.
        x2 (int): The X coordinate of the bottom-left point of the rectangle.
        y2 (int): The Y coordinate of the bottom-left point of the rectangle.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["boxColor"](renderer, x1, y1, x2, y2, color)

def boxRGBA(renderer, x1, y1, x2, y2, r, g, b, a):
    """Draws a filled rectangle to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the top-right point of the rectangle.
        y1 (int): The Y coordinate of the top-right point of the rectangle.
        x2 (int): The X coordinate of the bottom-left point of the rectangle.
        y2 (int): The Y coordinate of the bottom-left point of the rectangle.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["boxRGBA"](renderer, x1, y1, x2, y2, r, g, b, a)

def roundedBoxColor(renderer, x1, y1, x2, y2, rad, color):
    """Draws a filled rectangle with rounded corners to the renderer.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the top-right point of the rectangle.
        y1 (int): The Y coordinate of the top-right point of the rectangle.
        x2 (int): The X coordinate of the bottom-left point of the rectangle.
        y2 (int): The Y coordinate of the bottom-left point of the rectangle.
        rad (int): The radius of the arc of the rounded corners.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["roundedBoxColor"](renderer, x1, y1, x2, y2, rad, color)

def roundedBoxRGBA(renderer, x1, y1, x2, y2, rad, r, g, b, a):
    """Draws a filled rectangle with rounded corners to the renderer.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the top-right point of the rectangle.
        y1 (int): The Y coordinate of the top-right point of the rectangle.
        x2 (int): The X coordinate of the bottom-left point of the rectangle.
        y2 (int): The Y coordinate of the bottom-left point of the rectangle.
        rad (int): The radius of the arc of the rounded corners.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["roundedBoxRGBA"](renderer, x1, y1, x2, y2, rad, r, g, b, a)


def lineColor(renderer, x1, y1, x2, y2, color):
    """Draws a line to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the first point of the line.
        y1 (int): The Y coordinate of the first point of the line.
        x2 (int): The X coordinate of the second point of the line.
        y2 (int): The Y coordinate of the second point of the line.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["lineColor"](renderer, x1, y1, x2, y2, color)

def lineRGBA(renderer, x1, y1, x2, y2, r, g, b, a):
    """Draws a line to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the first point of the line.
        y1 (int): The Y coordinate of the first point of the line.
        x2 (int): The X coordinate of the second point of the line.
        y2 (int): The Y coordinate of the second point of the line.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["lineRGBA"](renderer, x1, y1, x2, y2, r, g, b, a)

def aalineColor(renderer, x1, y1, x2, y2, color):
    """Draws an anti-aliased line to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the first point of the line.
        y1 (int): The Y coordinate of the first point of the line.
        x2 (int): The X coordinate of the second point of the line.
        y2 (int): The Y coordinate of the second point of the line.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["aalineColor"](renderer, x1, y1, x2, y2, color)

def aalineRGBA(renderer, x1, y1, x2, y2, r, g, b, a):
    """Draws an anti-aliased line to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the first point of the line.
        y1 (int): The Y coordinate of the first point of the line.
        x2 (int): The X coordinate of the second point of the line.
        y2 (int): The Y coordinate of the second point of the line.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["aalineRGBA"](renderer, x1, y1, x2, y2, r, g, b, a)

def thickLineColor(renderer, x1, y1, x2, y2, width, color):
    """Draws a line with a given thickness to the renderer.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the first point of the line.
        y1 (int): The Y coordinate of the first point of the line.
        x2 (int): The X coordinate of the second point of the line.
        y2 (int): The Y coordinate of the second point of the line.
        width (int): The thickness of the line in pixels (from 1 to 255).
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["thickLineColor"](renderer, x1, y1, x2, y2, width, color)

def thickLineRGBA(renderer, x1, y1, x2, y2, width, r, g, b, a):
    """Draws a line with a given thickness to the renderer.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): The X coordinate of the first point of the line.
        y1 (int): The Y coordinate of the first point of the line.
        x2 (int): The X coordinate of the second point of the line.
        y2 (int): The Y coordinate of the second point of the line.
        width (int): The thickness of the line in pixels (from 1 to 255).
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["thickLineRGBA"](renderer, x1, y1, x2, y2, width, r, g, b, a)


def circleColor(renderer, x, y, rad, color):
    """Draws an unfilled circle to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the circle.
        y (int): The Y coordinate of the center of the circle.
        rad (int): The radius (in pixels) of the circle.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["circleColor"](renderer, x, y, rad, color)

def circleRGBA(renderer, x, y, rad, r, g, b, a):
    """Draws an unfilled circle to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the circle.
        y (int): The Y coordinate of the center of the circle.
        rad (int): The radius (in pixels) of the circle.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["circleRGBA"](renderer, x, y, rad, r, g, b, a)

def arcColor(renderer, x, y, rad, start, end, color):
    """Draws an arc to the renderer with a given color.

    The start and end of the arc are defined in units of degrees, with 0 being
    the bottom of the arc circle and increasing counter-clockwise (e.g. 90 being
    the rightmost point of the circle).

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the circle.
        y (int): The Y coordinate of the center of the circle.
        rad (int): The radius (in pixels) of the circle.
        start (int): The start of the arc (in degrees).
        end (int): The end of the arc (in degrees).
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["arcColor"](renderer, x, y, rad, start, end, color)

def arcRGBA(renderer, x, y, rad, start, end, r, g, b, a):
    """Draws an arc to the renderer with a given color.

    The start and end of the arc are defined in units of degrees, with 0 being
    the bottom of the arc circle and increasing counter-clockwise (e.g. 90 being
    the rightmost point of the circle).

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the circle.
        y (int): The Y coordinate of the center of the circle.
        rad (int): The radius (in pixels) of the circle.
        start (int): The start of the arc (in degrees).
        end (int): The end of the arc (in degrees).
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["arcRGBA"](renderer, x, y, rad, start, end, r, g, b, a)

def aacircleColor(renderer, x, y, rad, color):
    """Draws an anti-aliased unfilled circle to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the circle.
        y (int): The Y coordinate of the center of the circle.
        rad (int): The radius (in pixels) of the circle.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["aacircleColor"](renderer, x, y, rad, color)

def aacircleRGBA(renderer, x, y, rad, r, g, b, a):
    """Draws an anti-aliased unfilled circle to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the circle.
        y (int): The Y coordinate of the center of the circle.
        rad (int): The radius (in pixels) of the circle.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["aacircleRGBA"](renderer, x, y, rad, r, g, b, a)

def filledCircleColor(renderer, x, y, rad, color):
    """Draws a filled circle to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the circle.
        y (int): The Y coordinate of the center of the circle.
        rad (int): The radius (in pixels) of the circle.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["filledCircleColor"](renderer, x, y, rad, color)

def filledCircleRGBA(renderer, x, y, rad, r, g, b, a):
    """Draws a filled circle to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the circle.
        y (int): The Y coordinate of the center of the circle.
        rad (int): The radius (in pixels) of the circle.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["filledCircleRGBA"](renderer, x, y, rad, r, g, b, a)


def ellipseColor(renderer, x, y, rx, ry, color):
    """Draws an unfilled ellipse to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the ellipse.
        y (int): The Y coordinate of the center of the ellipse.
        rx (int): The x-axis radius (i.e. width) of the ellipse.
        ry (int): The y-axis radius (i.e. height) of the ellipse.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["ellipseColor"](renderer, x, y, rx, ry, color)

def ellipseRGBA(renderer, x, y, rx, ry, r, g, b, a):
    """Draws an unfilled ellipse to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the ellipse.
        y (int): The Y coordinate of the center of the ellipse.
        rx (int): The x-axis radius (i.e. width) of the ellipse.
        ry (int): The y-axis radius (i.e. height) of the ellipse.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["ellipseRGBA"](renderer, x, y, rx, ry, r, g, b, a)

def aaellipseColor(renderer, x, y, rx, ry, color):
    """Draws an anti-aliased unfilled ellipse to the renderer in a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the ellipse.
        y (int): The Y coordinate of the center of the ellipse.
        rx (int): The x-axis radius (i.e. width) of the ellipse.
        ry (int): The y-axis radius (i.e. height) of the ellipse.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["aaellipseColor"](renderer, x, y, rx, ry, color)

def aaellipseRGBA(renderer, x, y, rx, ry, r, g, b, a):
    """Draws an anti-aliased unfilled ellipse to the renderer in a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the ellipse.
        y (int): The Y coordinate of the center of the ellipse.
        rx (int): The x-axis radius (i.e. width) of the ellipse.
        ry (int): The y-axis radius (i.e. height) of the ellipse.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["aaellipseRGBA"](renderer, x, y, rx, ry, r, g, b, a)

def filledEllipseColor(renderer, x, y, rx, ry, color):
    """Draws a filled ellipse to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the ellipse.
        y (int): The Y coordinate of the center of the ellipse.
        rx (int): The x-axis radius (i.e. width) of the ellipse.
        ry (int): The y-axis radius (i.e. height) of the ellipse.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["filledEllipseColor"](renderer, x, y, rx, ry, color)

def filledEllipseRGBA(renderer, x, y, rx, ry, r, g, b, a):
    """Draws a filled ellipse to the renderer with a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the ellipse.
        y (int): The Y coordinate of the center of the ellipse.
        rx (int): The x-axis radius (i.e. width) of the ellipse.
        ry (int): The y-axis radius (i.e. height) of the ellipse.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["filledEllipseRGBA"](renderer, x, y, rx, ry, r, g, b, a)


def pieColor(renderer, x, y, rad, start, end, color):
    """Draws an unfilled pie slice (i.e. circle segment) to the renderer.

    The start and end of the pie are defined in units of degrees, with 0 being
    the bottom of the circle and increasing counter-clockwise (e.g. 90 being
    the rightmost point of the circle).

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the pie (circle).
        y (int): The Y coordinate of the center of the pie (circle).
        rad (int): The radius (in pixels) of the pie.
        start (int): Start of the pie slice (in degrees).
        end (int): End of the pie slice (in degrees)
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["pieColor"](renderer, x, y, rad, start, end, color)

def pieRGBA(renderer, x, y, rad, start, end, r, g, b, a):
    """Draws an unfilled pie slice (i.e. circle segment) to the renderer.

    The start and end of the pie are defined in units of degrees, with 0 being
    the bottom of the circle and increasing counter-clockwise (e.g. 90 being
    the rightmost point of the circle).

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the pie (circle).
        y (int): The Y coordinate of the center of the pie (circle).
        rad (int): The radius (in pixels) of the pie.
        start (int): Start of the pie slice (in degrees).
        end (int): End of the pie slice (in degrees)
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["pieRGBA"](renderer, x, y, rad, start, end, r, g, b, a)

def filledPieColor(renderer, x, y, rad, start, end, color):
    """Draws a filled pie slice (i.e. circle segment) to the renderer.

    The start and end of the pie are defined in units of degrees, with 0 being
    the bottom of the circle and increasing counter-clockwise (e.g. 90 being
    the rightmost point of the circle).

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the pie (circle).
        y (int): The Y coordinate of the center of the pie (circle).
        rad (int): The radius (in pixels) of the pie.
        start (int): Start of the pie slice (in degrees).
        end (int): End of the pie slice (in degrees)
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["filledPieColor"](renderer, x, y, rad, start, end, color)

def filledPieRGBA(renderer, x, y, rad, start, end, r, g, b, a):
    """Draws a filled pie slice (i.e. circle segment) to the renderer.

    The start and end of the pie are defined in units of degrees, with 0 being
    the bottom of the circle and increasing counter-clockwise (e.g. 90 being
    the rightmost point of the circle).

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the center of the pie (circle).
        y (int): The Y coordinate of the center of the pie (circle).
        rad (int): The radius (in pixels) of the pie.
        start (int): Start of the pie slice (in degrees).
        end (int): End of the pie slice (in degrees)
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["filledPieRGBA"](renderer, x, y, rad, start, end, r, g, b, a)


def trigonColor(renderer, x1, y1, x2, y2, x3, y3, color):
    """Draws a trigon (i.e. triangle outline) to the renderer in a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): X coordinate of the first point of the triangle.
        y1 (int): Y coordinate of the first point of the triangle.
        x2 (int): X coordinate of the second point of the triangle.
        y2 (int): Y coordinate of the second point of the triangle.
        x3 (int): X coordinate of the third point of the triangle.
        y3 (int): Y coordinate of the third point of the triangle.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["trigonColor"](renderer, x1, y1, x2, y2, x3, y3, color)

def trigonRGBA(renderer, x1, y1, x2, y2, x3, y3, r, g, b, a):
    """Draws a trigon (i.e. triangle outline) to the renderer in a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): X coordinate of the first point of the triangle.
        y1 (int): Y coordinate of the first point of the triangle.
        x2 (int): X coordinate of the second point of the triangle.
        y2 (int): Y coordinate of the second point of the triangle.
        x3 (int): X coordinate of the third point of the triangle.
        y3 (int): Y coordinate of the third point of the triangle.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["trigonRGBA"](renderer, x1, y1, x2, y2, x3, y3, r, g, b, a)

def aatrigonColor(renderer, x1, y1, x2, y2, x3, y3, color):
    """Draws an anti-aliased trigon (i.e. triangle outline) to the renderer.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): X coordinate of the first point of the triangle.
        y1 (int): Y coordinate of the first point of the triangle.
        x2 (int): X coordinate of the second point of the triangle.
        y2 (int): Y coordinate of the second point of the triangle.
        x3 (int): X coordinate of the third point of the triangle.
        y3 (int): Y coordinate of the third point of the triangle.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["aatrigonColor"](renderer, x1, y1, x2, y2, x3, y3, color)

def aatrigonRGBA(renderer, x1, y1, x2, y2, x3, y3, r, g, b, a):
    """Draws an anti-aliased trigon (i.e. triangle outline) to the renderer.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): X coordinate of the first point of the triangle.
        y1 (int): Y coordinate of the first point of the triangle.
        x2 (int): X coordinate of the second point of the triangle.
        y2 (int): Y coordinate of the second point of the triangle.
        x3 (int): X coordinate of the third point of the triangle.
        y3 (int): Y coordinate of the third point of the triangle.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["aatrigonRGBA"](renderer, x1, y1, x2, y2, x3, y3, r, g, b, a)

def filledTrigonColor(renderer, x1, y1, x2, y2, x3, y3, color):
    """Draws a filled triangle to the renderer in a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): X coordinate of the first point of the triangle.
        y1 (int): Y coordinate of the first point of the triangle.
        x2 (int): X coordinate of the second point of the triangle.
        y2 (int): Y coordinate of the second point of the triangle.
        x3 (int): X coordinate of the third point of the triangle.
        y3 (int): Y coordinate of the third point of the triangle.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["filledTrigonColor"](renderer, x1, y1, x2, y2, x3, y3, color)

def filledTrigonRGBA(renderer, x1, y1, x2, y2, x3, y3, r, g, b, a):
    """Draws a filled triangle to the renderer in a given color.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x1 (int): X coordinate of the first point of the triangle.
        y1 (int): Y coordinate of the first point of the triangle.
        x2 (int): X coordinate of the second point of the triangle.
        y2 (int): Y coordinate of the second point of the triangle.
        x3 (int): X coordinate of the third point of the triangle.
        y3 (int): Y coordinate of the third point of the triangle.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["filledTrigonRGBA"](renderer, x1, y1, x2, y2, x3, y3, r, g, b, a)


def polygonColor(renderer, vx, vy, n, color):
    """Draws an unfilled polygon to the renderer in a given color.

    Vertices are specified as ``ctypes.c_int16`` arrays, with two arrays of
    equal size defining the x and y coordinates of the points making up the
    polygon. To create these vertex arrays in Python, you can create lists and
    cast them to ctypes arrays which can be passed directly to the function::

       x_coords = [5, 5, 15, 15]
       y_coords = [5, 10, 10, 5]
       vx = (ctypes.c_int16 * len(x_coords))(*x_coords)
       vy = (ctypes.c_int16 * len(y_coords))(*y_coords)

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the X coordinates
            of the polygon's vertices.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the Y coordinates
            of the polygon's vertices.
        n (int): The number of vertices in the polygon.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["polygonColor"](renderer, vx, vy, n, color)

def polygonRGBA(renderer, vx, vy, n, r, g, b, a):
    """Draws an unfilled polygon to the renderer in a given color.

    See :func:`polygonColor` for more information on usage.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the X coordinates
            of the polygon's vertices.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the Y coordinates
            of the polygon's vertices.
        n (int): The number of vertices in the polygon.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["polygonRGBA"](renderer, vx, vy, n, r, g, b, a)

def aapolygonColor(renderer, vx, vy, n, color):
    """Draws an anti-aliased unfilled polygon to the renderer in a given color.

    See :func:`polygonColor` for more information on usage.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the X coordinates
            of the polygon's vertices.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the Y coordinates
            of the polygon's vertices.
        n (int): The number of vertices in the polygon.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["aapolygonColor"](renderer, vx, vy, n, color)

def aapolygonRGBA(renderer, vx, vy, n, r, g, b, a):
    """Draws an anti-aliased unfilled polygon to the renderer in a given color.

    See :func:`polygonColor` for more information on usage.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the X coordinates
            of the polygon's vertices.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the Y coordinates
            of the polygon's vertices.
        n (int): The number of vertices in the polygon.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["aapolygonRGBA"](renderer, vx, vy, n, r, g, b, a)

def filledPolygonColor(renderer, vx, vy, n, color):
    """Draws a filled polygon to the renderer in a given color.

    See :func:`polygonColor` for more information on usage.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the X coordinates
            of the polygon's vertices.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the Y coordinates
            of the polygon's vertices.
        n (int): The number of vertices in the polygon.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["filledPolygonColor"](renderer, vx, vy, n, color)

def filledPolygonRGBA(renderer, vx, vy, n, r, g, b, a):
    """Draws a filled polygon to the renderer in a given color.

    See :func:`polygonColor` for more information on usage.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the X coordinates
            of the polygon's vertices.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the Y coordinates
            of the polygon's vertices.
        n (int): The number of vertices in the polygon.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["filledPolygonRGBA"](renderer, vx, vy, n, r, g, b, a)

def texturedPolygon(renderer, vx, vy, n, texture, texture_dx, texture_dy):
    """Draws a polygon to the renderer with a given texture.

    The location of the texture is relative to the top-left corner of the
    renderer, as opposed to being relative to the polygon itself. As such,
    both the vertex coordinates and texture coordinates need to be adjusted
    equally to render a polygon with the same texture placement at a different
    location.

    The texture must be associated with the same renderer used to draw the
    polygon.

    See :func:`polygonColor` for more information on usage.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the X coordinates
            of the polygon's vertices.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the Y coordinates
            of the polygon's vertices.
        n (int): The number of vertices in the polygon.
        texture (:obj:`SDL_Texture`): The texture with which to fill the
            polygon.
        texture_dx (int): The X offset of the texture relative to the top-left
            corner of the renderer.
        texture_dy (int): The Y offset of the texture relative to the top-left
            corner of the renderer.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["texturedPolygon"](
        renderer, vx, vy, n, texture, texture_dx, texture_dy
    )


def bezierColor(renderer, vx, vy, n, s, color):
    """Draws a Bezier curve to the renderer in a given color.

    The first and last vertex are the start and end points of the Bezier,
    respectively, with the points in between defining the control points of the
    curve. For example, a 3rd order (i.e. cubic) Bezier would be defined using
    4 vertices, with the two middle vertices being the control points.

    See :func:`polygonColor` for more information on creating the vertex arrays
    for this function.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the X coordinates
            of the points of the Bezier curve.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the Y coordinates
            of the points of the Bezier curve.
        n (int): The number of points in the bezier curve (minimum of 3).
        s (int): The number of interpolation steps to use when drawing the
            curve (minimum of 2). The higher the value, the smoother the curve.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["bezierColor"](renderer, vx, vy, n, s, color)

def bezierRGBA(renderer, vx, vy, n, s, r, g, b, a):
    """Draws a Bezier curve to the renderer in a given color.

    See :func:`bezierColor` for more details on usage.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the X coordinates
            of the points of the Bezier curve.
        vx (POINTER(:obj:`~ctypes.c_int16`)): Array containing the Y coordinates
            of the points of the Bezier curve.
        n (int): The number of points in the bezier curve (minimum of 3).
        s (int): The number of interpolation steps to use when drawing the
            curve (minimum of 2). The higher the value, the smoother the curve.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["bezierRGBA"](renderer, vx, vy, n, s, r, g, b, a)


def gfxPrimitivesSetFont(fontdata, cw, ch):
    """Sets or resets the current global GFX font.

    The SDL_gfx library uses its own special format for bitmap fonts. Basically,
    fonts are byte arrays where each glyph is made up of the same number of
    bytes (as defined by the ``cw`` and ``ch`` arguments). The bytes are used as
    a binary bitmask with 1s indicating the pixels of the character and 0s
    indicating the transparent background. For example, the following is the
    definition of the capital H glyph in the default font:

    .. code-block:: c

       /*
       * 72 0x48 'H' 
       */
       0xc6,			/* 11000110 */
       0xc6,			/* 11000110 */
       0xc6,			/* 11000110 */
       0xfe,			/* 11111110 */
       0xc6,			/* 11000110 */
       0xc6,			/* 11000110 */
       0xc6,			/* 11000110 */
       0x00,			/* 00000000 */

    Each font must contain glyphs for all 256 ASCII characters. Since this is
    a pretty painful format for defining your own fonts, you can load and use
    any of the predefined SDL_gfx fonts from the following link:
    https://github.com/ferzkopp/SDL_gfx/tree/master/Fonts

    If no font has been set, SDL_gfx defaults to rendering with a built-in 8x8
    pixel font.

    .. note::
        If anyone comes up with a way of converting standard bitmap fonts into
        the SDL_gfx format, please let us know! That would be incredibly cool
        and handy.

    Args:
        fontdata (:obj:`~ctypes.c_void_p`): A pointer to the start of the array
            containing the new global font data, or a null pointer to reset the
            global font to the default 8x8 font.
        cw (int): The width (in bytes) of each character of the font. Ignored if
            ``fontdata`` is null.
        ch (int): The height (in bytes) of each character of the font. Ignored
            if ``fontdata`` is null.

    """
    return _funcs["gfxPrimitivesSetFont"](fontdata, cw, ch)

def gfxPrimitivesSetFontRotation(rotation):
    """Sets the global character rotation for GFX font rendering.

    Characters can only be rotated in 90 degree increments. Calling this
    function will reset the character cache.

    Args:
        rotation (int): The number of clockwise 90-degree rotations to apply to
            font characters when rendering text.

    """
    return _funcs["gfxPrimitivesSetFontRotation"](rotation)

def characterColor(renderer, x, y, c, color):
    """Draws a single character with the current GFX font to the renderer.

    Python characters can be converted to ASCII integers for use with this
    function using the built-in :func:`ord` function (e.g. ``ord(u"A")``).

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the upper-left corner of the character.
        y (int): The Y coordinate of the upper-left corner of the character.
        c (int): The ASCII number (from 0 to 255) of the character.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["characterColor"](renderer, x, y, c, color)

def characterRGBA(renderer, x, y, c, r, g, b, a):
    """Draws a single character with the current GFX font to the renderer.

    See :func:`characterColor` for more usage information.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the upper-left corner of the character.
        y (int): The Y coordinate of the upper-left corner of the character.
        c (int): The ASCII number (from 0 to 255) of the character.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["characterRGBA"](renderer, x, y, c, r, g, b, a)

def stringColor(renderer, x, y, s, color):
    """Draws an ASCII string with the current GFX font to the renderer.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the upper-left corner of the string.
        y (int): The Y coordinate of the upper-left corner of the string.
        s (bytes): The ASCII-encoded bytestring of text to render.
        color (int): The color to draw with as a 32-bit ``0xRRGGBBAA`` integer
            (e.g. ``0xFF0000FF`` for solid red).

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["stringColor"](renderer, x, y, s, color)

def stringRGBA(renderer, x, y, s, r, g, b, a):
    """Draws an ASCII string with the current GFX font to the renderer.

    If the rendering color has any transparency, blending will be enabled.

    Args:
        renderer (:obj:`SDL_Renderer`): The renderer to draw on.
        x (int): The X coordinate of the upper-left corner of the string.
        y (int): The Y coordinate of the upper-left corner of the string.
        s (bytes): The ASCII-encoded bytestring of text to render.
        r (int): The red value (from 0 to 255) of the color to draw with.
        g (int): The green value (from 0 to 255) of the color to draw with.
        b (int): The blue value (from 0 to 255) of the color to draw with.
        a (int): The alpha value (from 0 to 255) of the color to draw with.

    Returns:
        int: 0 on success, or -1 on failure.

    """
    return _funcs["stringRGBA"](renderer, x, y, s, r, g, b, a)


def rotozoomSurface(src, angle, zoom, smooth):
    """Rotates & zooms a surface.

    Rotates and zooms an :obj:`SDL_Surface` to a new output surface, with
    optional anti-aliasing. If the surface is not 8-bit or 32-bit RGBA/ABGR, it
    will be converted into a 32-bit RGBA format on the fly.
    
    Args:
        src (:obj:`SDL_Surface`): The surface to rotate and zoom.
        angle (float): The angle to rotate the surface (in degrees).
        zoom (float): The scaling factor for the surface.
        smooth (int): If set to 1, the output image will be anti-aliased. If set
            to 0, no anti-aliasing will be performed. Must be either 0 or 1.
    
    Returns:
        :obj:`SDL_Surface`: A new output surface with zoom & rotation applied.

    """
    return _funcs["rotozoomSurface"](src, angle, zoom, smooth)

def rotozoomSurfaceXY(src, angle, zoomx, zoomy, smooth):
    """Rotates & zooms a surface with different x & y scaling factors.

    Rotates and zooms an :obj:`SDL_Surface` to a new output surface, with
    optional anti-aliasing. If the surface is not 8-bit or 32-bit RGBA/ABGR, it
    will be converted into a 32-bit RGBA format on the fly.
    
    Args:
        src (:obj:`SDL_Surface`): The surface to rotate and zoom.
        angle (float): The angle to rotate the surface (in degrees).
        zoomx (float): The x-axis (horizontal) scaling factor.
        zoomy (float): The y-axis (vertical) scaling factor.
        smooth (int): If set to 1, the output image will be anti-aliased. If set
            to 0, no anti-aliasing will be performed. Must be either 0 or 1.
    
    Returns:
        :obj:`SDL_Surface`: A new output surface with zoom & rotation applied.

    """
    return _funcs["rotozoomSurfaceXY"](src, angle, zoomx, zoomy, smooth)

def rotozoomSurfaceSize(width, height, angle, zoom, dstwidth, dstheight):
    """Returns the output surface size of a :func:`rotozoomSurface` call.

    This function outputs the calculated height and width by reference to the
    ``dstwidth`` and ``dstheight`` arguments, and does not return any value
    itself.

    Args:
        width (int): The width (in pixels) of the source surface.
        height (int): The height (in pixels) of the source surface.
        angle (float): The angle to rotate the surface (in degrees).
        zoom (float): The scaling factor for the surface.
        dstwidth (byref(`c_int`)): A reference to the ctypes int where the
            calculated width of the output surface will be stored.
        dstheight (byref(`c_int`)): A reference to the ctypes int where the
            calculated height of the output surface will be stored.

    """
    return _funcs["rotozoomSurfaceSize"](
        width, height, angle, zoom, dstwidth, dstheight
    )

def rotozoomSurfaceSizeXY(width, height, angle, zoomx, zoomy, dstwidth, dstheight):
    """Returns the output surface size of a :func:`rotozoomSurfaceXY` call.

    This function outputs the calculated height and width by reference to the
    ``dstwidth`` and ``dstheight`` arguments, and does not return any value
    itself.

    Args:
        width (int): The width (in pixels) of the source surface.
        height (int): The height (in pixels) of the source surface.
        angle (float): The angle to rotate the surface (in degrees).
        zoomx (float): The x-axis (horizontal) scaling factor.
        zoomy (float): The y-axis (vertical) scaling factor.
        dstwidth (byref(`c_int`)): A reference to the ctypes int where the
            calculated width of the output surface will be stored.
        dstheight (byref(`c_int`)): A reference to the ctypes int where the
            calculated height of the output surface will be stored.

    """
    return _funcs["rotozoomSurfaceSizeXY"](
        width, height, angle, zoomx, zoomy, dstwidth, dstheight
    )

def zoomSurface(src, zoomx, zoomy, smooth):
    """Zooms a surface with different x & y scaling factors.

    This function renders to a new surface, with optional anti-aliasing. If a
    zoom factor is negative, the image will be flipped along that axis. If the
    surface is not 8-bit or 32-bit RGBA/ABGR, it will be converted into a 32-bit
    RGBA format on the fly.

    Args:
        src (:obj:`SDL_Surface`): The surface to zoom.
        zoomx (float): The x-axis (horizontal) zoom factor.
        zoomy (float): The y-axis (vertical) zoom factor.
        smooth (int): If set to 1, the output image will be anti-aliased. If set
            to 0, no anti-aliasing will be performed. Must be either 0 or 1.
    
    Returns:
        :obj:`SDL_Surface`: A new output surface with zoom applied.

    """
    return _funcs["zoomSurface"](src, zoomx, zoomy, smooth)

def zoomSurfaceSize(width, height, zoomx, zoomy, dstwidth, dstheight):
    """Returns the output surface size of a :func:`zoomSurface` call.

    This function outputs the calculated height and width by reference to the
    ``dstwidth`` and ``dstheight`` arguments, and does not return any value
    itself.

    Args:
        width (int): The width (in pixels) of the source surface.
        height (int): The height (in pixels) of the source surface.
        zoomx (float): The x-axis (horizontal) scaling factor.
        zoomy (float): The y-axis (vertical) scaling factor.
        dstwidth (byref(`c_int`)): A reference to the ctypes int where the
            calculated width of the output surface will be stored.
        dstheight (byref(`c_int`)): A reference to the ctypes int where the
            calculated height of the output surface will be stored.

    """
    return _funcs["zoomSurfaceSize"](width, height, zoomx, zoomy, dstwidth, dstheight)

def shrinkSurface(src, factorx, factory):
    """Shrinks a surface by an integer ratio using averaging.

    This function renders to a new surface, meaning that the original surface is
    not modified. The output surface is anti-aliased by averaging the source
    RGBA information. If the surface is not 8-bit or 32-bit RGBA/ABGR, it will
    be converted into a 32-bit RGBA format on the fly.

    Args:
        src (:obj:`SDL_Surface`): The surface to zoom.
        factorx (int): The x-axis (horizontal) shrink factor (e.g. 2 = 2x smaller).
        factory (int): The y-axis (vertical) shrink factor (e.g. 2 = 2x smaller).


    Returns:
        :obj:`SDL_Surface`: The new shrunken surface.

    """
    return _funcs["shrinkSurface"](src, factorx, factory)

def rotateSurface90Degrees(src, numClockwiseTurns):
    """Rotates an SDL surface clockwise in increments of 90 degrees.

    Faster than rotozoomer since no scanning or interpolation takes place.
    Input surface must be 8-bit, 16-bit, 24-bit, or 32-bit.

    Args:
        src (:obj:`SDL_Surface`): The source surface to rotate.
        numClockwiseTurns (int): The number of clockwise 90 degree rotations to
            apply to the source.

    Returns:
        :obj:`SDL_Surface`: The new rotated surface, or `None` if the source
        surface was not a compatible format.

    """
    return _funcs["rotateSurface90Degrees"](src, numClockwiseTurns)



import pytest
from sdl2 import SDL_Init, SDL_Quit, SDL_INIT_VIDEO
from sdl2 import vulkan


@pytest.mark.skip("not implemented")
def test_SDL_Vulkan_LoadLibrary(with_sdl):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_Vulkan_GetVkGetInstanceProcAddr(with_sdl):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_Vulkan_UnloadLibrary(with_sdl):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_Vulkan_GetInstanceExtensions(with_sdl):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_Vulkan_CreateSurface(with_sdl):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_Vulkan_GetDrawableSize(with_sdl):
    pass



import os
import sys
import ctypes
import sdl2
from sdl2 import SDL_Init, SDL_Quit, SDL_InitSubSystem, SDL_QuitSubSystem, \
    SDL_INIT_AUDIO
from sdl2.error import SDL_GetError, SDL_ClearError
from sdl2 import audio
import pytest

# NOTE: This module is missing a lot of tests, but is also going to be tricky
# to write more tests for.

# Get original audio driver, if one was set in the environment
original_driver = os.getenv("SDL_AUDIODRIVER", None)

@pytest.fixture
def with_sdl_audio():
    # Reset original audio driver in the environment (if there was one)
    if original_driver:
        os.environ["SDL_AUDIODRIVER"] = original_driver
    # Initialize SDL2 with video and audio subsystems
    sdl2.SDL_ClearError()
    ret = sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO | sdl2.SDL_INIT_AUDIO)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    yield
    sdl2.SDL_Quit()
    # Reset original audio driver in environment
    os.environ.pop("SDL_AUDIODRIVER", None)
    if original_driver:
        os.environ["SDL_AUDIODRIVER"] = original_driver


# Test macro functions

def test_SDL_AUDIO_BITSIZE():
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_U8) == 8
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_S8) == 8
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_U16LSB) == 16
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_S16LSB) == 16
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_U16MSB) == 16
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_S16MSB) == 16
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_U16) == 16
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_S16) == 16
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_S32LSB) == 32
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_S32MSB) == 32
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_S32) == 32
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_F32LSB) == 32
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_F32MSB) == 32
    assert audio.SDL_AUDIO_BITSIZE(audio.AUDIO_F32) == 32

def test_SDL_AUDIO_ISFLOAT():
    assert not audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_U8)
    assert not audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_S8)
    assert not audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_U16LSB)
    assert not audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_S16LSB)
    assert not audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_U16MSB)
    assert not audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_S16MSB)
    assert not audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_U16)
    assert not audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_S16)
    assert not audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_S32LSB)
    assert not audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_S32MSB)
    assert not audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_S32)
    assert audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_F32LSB)
    assert audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_F32MSB)
    assert audio.SDL_AUDIO_ISFLOAT(audio.AUDIO_F32)

def test_SDL_AUDIO_ISBIGENDIAN():
    assert not audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_U8)
    assert not audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_S8)
    assert not audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_U16LSB)
    assert not audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_S16LSB)
    assert audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_U16MSB)
    assert audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_S16MSB)
    assert not audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_U16)
    assert not audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_S16)
    assert not audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_S32LSB)
    assert audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_S32MSB)
    assert not audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_S32)
    assert not audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_F32LSB)
    assert audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_F32MSB)
    assert not audio.SDL_AUDIO_ISBIGENDIAN(audio.AUDIO_F32)

def test_SDL_AUDIO_ISSIGNED():
    assert not audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_U8)
    assert audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_S8)
    assert not audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_U16LSB)
    assert audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_S16LSB)
    assert not audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_U16MSB)
    assert audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_S16MSB)
    assert not audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_U16)
    assert audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_S16)
    assert audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_S32LSB)
    assert audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_S32MSB)
    assert audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_S32)
    assert audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_F32LSB)
    assert audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_F32MSB)
    assert audio.SDL_AUDIO_ISSIGNED(audio.AUDIO_F32)

def test_SDL_AUDIO_ISINT():
    assert audio.SDL_AUDIO_ISINT(audio.AUDIO_U8)
    assert audio.SDL_AUDIO_ISINT(audio.AUDIO_S8)
    assert audio.SDL_AUDIO_ISINT(audio.AUDIO_U16LSB)
    assert audio.SDL_AUDIO_ISINT(audio.AUDIO_S16LSB)
    assert audio.SDL_AUDIO_ISINT(audio.AUDIO_U16MSB)
    assert audio.SDL_AUDIO_ISINT(audio.AUDIO_S16MSB)
    assert audio.SDL_AUDIO_ISINT(audio.AUDIO_U16)
    assert audio.SDL_AUDIO_ISINT(audio.AUDIO_S16)
    assert audio.SDL_AUDIO_ISINT(audio.AUDIO_S32LSB)
    assert audio.SDL_AUDIO_ISINT(audio.AUDIO_S32MSB)
    assert audio.SDL_AUDIO_ISINT(audio.AUDIO_S32)
    assert not audio.SDL_AUDIO_ISINT(audio.AUDIO_F32LSB)
    assert not audio.SDL_AUDIO_ISINT(audio.AUDIO_F32MSB)
    assert not audio.SDL_AUDIO_ISINT(audio.AUDIO_F32)

def test_SDL_AUDIO_ISLITTLEENDIAN():
    assert audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_U8)
    assert audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_S8)
    assert audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_U16LSB)
    assert audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_S16LSB)
    assert not audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_U16MSB)
    assert not audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_S16MSB)
    assert audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_U16)
    assert audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_S16)
    assert audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_S32LSB)
    assert not audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_S32MSB)
    assert audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_S32)
    assert audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_F32LSB)
    assert not audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_F32MSB)
    assert audio.SDL_AUDIO_ISLITTLEENDIAN(audio.AUDIO_F32)

def test_SDL_AUDIO_ISUNSIGNED():
    assert audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_U8)
    assert not audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_S8)
    assert audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_U16LSB)
    assert not audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_S16LSB)
    assert audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_U16MSB)
    assert not audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_S16MSB)
    assert audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_U16)
    assert not audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_S16)
    assert not audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_S32LSB)
    assert not audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_S32MSB)
    assert not audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_S32)
    assert not audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_F32LSB)
    assert not audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_F32MSB)
    assert not audio.SDL_AUDIO_ISUNSIGNED(audio.AUDIO_F32)


# Test structs and objects

@pytest.mark.skip("not implemented")
def test_SDL_AudioSpec():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_AudioCVT():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_AudioStream():
    pass


# Test actual function bindings

@pytest.mark.skip("not implemented")
def test_SDL_AudioInitQuit():
    pass

def test_SDL_GetNumAudioDrivers(with_sdl_audio):
    count = audio.SDL_GetNumAudioDrivers()
    assert count >= 1

def test_SDL_GetAudioDriver(with_sdl_audio):
    founddummy = False
    drivercount = audio.SDL_GetNumAudioDrivers()
    for index in range(drivercount):
        drivername = audio.SDL_GetAudioDriver(index)
        assert isinstance(drivername, (str, bytes))
        if drivername == b"dummy":
            founddummy = True
    assert founddummy

def test_SDL_GetCurrentAudioDriver(with_sdl_audio):
    success = 0
    # Reset audio subsystem
    SDL_Quit()
    SDL_Init(0)
    for index in range(audio.SDL_GetNumAudioDrivers()):
        drivername = audio.SDL_GetAudioDriver(index)
        os.environ["SDL_AUDIODRIVER"] = drivername.decode("utf-8")
        # Certain drivers fail without bringing up the correct
        # return value, such as the esd, if it is not running.
        SDL_InitSubSystem(SDL_INIT_AUDIO)
        driver = audio.SDL_GetCurrentAudioDriver()
        SDL_QuitSubSystem(SDL_INIT_AUDIO)
        # Do not handle wrong return values.
        if driver is not None:
            assert drivername == driver
            success += 1
    assert success >= 1

def test_SDL_OpenCloseAudio(with_sdl_audio):
    # TODO: Add test that checks which audio formats are supported for each
    # audio device?
    fmt = audio.AUDIO_F32 if sys.platform == "darwin" else audio.AUDIO_U16
    reqspec = audio.SDL_AudioSpec(44100, fmt, 2, 1024)
    spec = audio.SDL_AudioSpec(0, 0, 0, 0)
    ret = audio.SDL_OpenAudio(reqspec, ctypes.byref(spec))
    assert ret == 0
    assert spec.format > 0  # Can't guarantee we'll get requested format
    assert spec.freq == reqspec.freq
    assert spec.channels == reqspec.channels
    audio.SDL_CloseAudio()

def test_SDL_GetNumAudioDevices(with_sdl_audio):
    outnum = audio.SDL_GetNumAudioDevices(False)
    assert outnum >= 1
    innum = audio.SDL_GetNumAudioDevices(True)
    assert innum >= 0

def test_SDL_GetAudioDeviceName(with_sdl_audio):
    # NOTE: Check & print errors for drivers that failed to load?
    backends = []
    devices = {}
    # Reset audio subsystem
    SDL_Quit()
    SDL_Init(0)
    for index in range(audio.SDL_GetNumAudioDrivers()):
        # Get input/output device names for each audio driver
        drivername = audio.SDL_GetAudioDriver(index)
        backends.append(drivername.decode("utf-8"))
        os.environ["SDL_AUDIODRIVER"] = drivername.decode("utf-8")
        # Need to reinitialize subsystem for each driver
        SDL_InitSubSystem(SDL_INIT_AUDIO)
        driver = audio.SDL_GetCurrentAudioDriver()
        if driver is not None:
            driver = driver.decode("utf-8")
            devices[driver] = {'input': [], 'output': []}
            outnum = audio.SDL_GetNumAudioDevices(False)
            innum = audio.SDL_GetNumAudioDevices(True)
            for x in range(outnum):
                name = audio.SDL_GetAudioDeviceName(x, False)
                assert name is not None
                devices[driver]['output'].append(name.decode('utf-8'))
            for x in range(innum):
                name = audio.SDL_GetAudioDeviceName(x, True)
                assert name is not None
                devices[driver]['input'].append(name.decode('utf-8'))
        SDL_QuitSubSystem(SDL_INIT_AUDIO)
    print("Audio backends supported by current SDL2 binary:")
    print(backends)
    print("\nAvailable audio drivers and devices:")
    for driver in devices.keys():
        print(driver)
        print(" - input: {0}".format(str(devices[driver]['input'])))
        print(" - output: {0}".format(str(devices[driver]['output'])))

@pytest.mark.skipif(sdl2.dll.version < 2016, reason="not available")
def test_SDL_GetAudioDeviceSpec(with_sdl_audio):
    # Reset audio subsystem
    SDL_Quit()
    SDL_Init(0)
    # Find an audio driver with at least one output
    SDL_InitSubSystem(SDL_INIT_AUDIO)
    driver = audio.SDL_GetCurrentAudioDriver()
    if driver == None or audio.SDL_GetNumAudioDevices(False) == 0:
        SDL_QuitSubSystem(SDL_INIT_AUDIO)
        os.environ["SDL_AUDIODRIVER"] = b'dummy'
        SDL_InitSubSystem(SDL_INIT_AUDIO)
        driver = audio.SDL_GetCurrentAudioDriver()
    drivername = driver.decode('utf-8')
    # Get name and spec of first output device
    outspec = audio.SDL_AudioSpec(0, 0, 0, 0)
    outname = audio.SDL_GetAudioDeviceName(0, False).decode('utf-8')
    ret = audio.SDL_GetAudioDeviceSpec(0, False, ctypes.byref(outspec))
    SDL_QuitSubSystem(SDL_INIT_AUDIO)
    assert ret == 0
    # Validate frequency and channel count were set
    hz = outspec.freq
    fmt = audio.FORMAT_NAME_MAP[outspec.format] if outspec.format > 0 else 'unknown'
    chans = outspec.channels
    bufsize = outspec.samples if outspec.samples > 0 else 'unknown'
    if driver != b"dummy":
        assert hz > 0
        assert chans > 0
        # Print out device spec info
        msg = "Audio device spec for {0} with '{1}' driver:"
        msg2 = "{0} Hz, {1} channels, {2} format, {3} sample buffer size"
        print(msg.format(outname, drivername))
        print(msg2.format(hz, chans, fmt, bufsize))

def test_SDL_OpenCloseAudioDevice(with_sdl_audio):
    #TODO: Add tests for callback
    fmt = audio.AUDIO_F32 if sys.platform == "darwin" else audio.AUDIO_U16
    reqspec = audio.SDL_AudioSpec(44100, fmt, 2, 1024)
    outnum = audio.SDL_GetNumAudioDevices(0)
    for x in range(outnum):
        spec = audio.SDL_AudioSpec(0, 0, 0, 0)
        name = audio.SDL_GetAudioDeviceName(x, 0)
        assert name is not None
        deviceid = audio.SDL_OpenAudioDevice(
            name, 0, reqspec, ctypes.byref(spec),
            audio.SDL_AUDIO_ALLOW_ANY_CHANGE
        )
        err = SDL_GetError()
        assert deviceid >= 2
        assert isinstance(spec, audio.SDL_AudioSpec)
        assert spec.format in audio.AUDIO_FORMATS
        assert spec.freq > 0
        assert spec.channels > 0
        assert spec.samples > 0
        audio.SDL_CloseAudioDevice(deviceid)

@pytest.mark.skip("not implemented")
def test_SDL_GetAudioStatus(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GetAudioDeviceStatus(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_PauseAudio(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_PauseAudioDevice(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_LoadWAV_RW(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_LoadWAV(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_FreeWAV(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_BuildAudioCVT(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_ConvertAudio(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_MixAudio(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_MixAudioFormat(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_LockUnlockAudio(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_LockUnlockAudioDevice(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_QueueAudio(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GetQueuedAudioSize(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_ClearQueuedAudio(self):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_DequeueAudio(self):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2007, reason="not available")
def test_SDL_NewAudioStream():
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2007, reason="not available")
def test_SDL_AudioStreamPut():
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2007, reason="not available")
def test_SDL_AudioStreamGet():
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2007, reason="not available")
def test_SDL_AudioStreamAvailable():
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2007, reason="not available")
def test_SDL_AudioStreamClear():
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2007, reason="not available")
def test_SDL_FreeAudioStream():
    pass



import sys
import sdl2
import sdl2.ext

# Create a resource container, so that we can easily access all
# resource, we bundle with our application. We are using the current
# file's location and define the "resources" subdirectory as the
# location, in which we keep all data.
RESOURCES = sdl2.ext.Resources(__file__, "resources")

def run():
    # Initialize the video subsystem, create a window and make it visible.
    sdl2.ext.init()
    window = sdl2.ext.Window("Sprite Transformations", size=(800, 600))
    window.show()

    # Create a hardware-accelerated renderer for drawing to the new window.
    # We'll also set the default scaling quality to 'best' for smoother
    # edges when rendering.
    renderer = sdl2.ext.Renderer(window)
    sdl2.ext.set_texture_scale_quality(method="best")

    # Import an image file and convert it to a Texture for the renderer
    tst_img = sdl2.ext.load_bmp(RESOURCES.get_path("hello.bmp"))
    tx = sdl2.ext.Texture(renderer, tst_img)
    
    angle = 0
    flip = 0
    running = True
    while running:
        events = sdl2.ext.get_events()
        for event in events:
            if event.type == sdl2.SDL_QUIT:
                running = False
                break
            elif event.type == sdl2.SDL_KEYDOWN:
                # Flip the sprite vertically using the up/down arrow keys
                if event.key.keysym.sym in (sdl2.SDLK_DOWN, sdl2.SDLK_UP):
                    flip ^= sdl2.SDL_FLIP_VERTICAL
                # Flip the sprite horizontally using the left/right arrow keys
                elif event.key.keysym.sym in (sdl2.SDLK_LEFT, sdl2.SDLK_RIGHT):
                    flip ^= sdl2.SDL_FLIP_HORIZONTAL
                # Rotate the sprite around its center using the (+/-) keys
                elif event.key.keysym.sym == sdl2.SDLK_EQUALS:
                    angle += 1.0
                    if angle >= 360.0:
                        angle = 0.0
                elif event.key.keysym.sym == sdl2.SDLK_MINUS:
                    angle -= 1.0
                    if angle <= 0.0:
                        angle = 360.0

        # Clear the renderer, copy our texture to it at an offset of (100, 75)
        # from the top-left corner, present it to the window, and wait 10 ms 
        # before moving on
        renderer.clear()
        renderer.copy(tx, dstrect=(100, 75), angle=angle, flip=flip)
        renderer.present()
        sdl2.SDL_Delay(10)

    sdl2.ext.quit()
    return 0

if __name__ == "__main__":
    sys.exit(run())



import sys
from ctypes import (Structure, Union, c_int, c_char_p, c_void_p, c_float,
    POINTER, create_string_buffer)
from .dll import _bind
from .dll import version as sdl_version
from .stdinc import SDL_bool, Sint16, Uint32, Uint16, Uint8
from .joystick import (SDL_JoystickGUID, SDL_Joystick, SDL_JoystickID,
    SDL_JoystickGetGUIDString)
from .rwops import SDL_RWops, SDL_RWFromFile
from .sensor import SDL_SensorType

__all__ = [
    # Structs & Opaque Types
    "SDL_GameController", "SDL_GameControllerButtonBind",

    # Enums
    "SDL_GameControllerType",
    "SDL_CONTROLLER_TYPE_UNKNOWN", "SDL_CONTROLLER_TYPE_XBOX360",
    "SDL_CONTROLLER_TYPE_XBOXONE", "SDL_CONTROLLER_TYPE_PS3",
    "SDL_CONTROLLER_TYPE_PS4", "SDL_CONTROLLER_TYPE_NINTENDO_SWITCH_PRO",
    "SDL_CONTROLLER_TYPE_VIRTUAL", "SDL_CONTROLLER_TYPE_PS5",
    "SDL_CONTROLLER_TYPE_AMAZON_LUNA", "SDL_CONTROLLER_TYPE_GOOGLE_STADIA",

    "SDL_GameControllerBindType",
    "SDL_CONTROLLER_BINDTYPE_NONE", "SDL_CONTROLLER_BINDTYPE_BUTTON",
    "SDL_CONTROLLER_BINDTYPE_AXIS", "SDL_CONTROLLER_BINDTYPE_HAT", 

    "SDL_GameControllerAxis",
    "SDL_CONTROLLER_AXIS_INVALID", "SDL_CONTROLLER_AXIS_LEFTX",
    "SDL_CONTROLLER_AXIS_LEFTY", "SDL_CONTROLLER_AXIS_RIGHTX",
    "SDL_CONTROLLER_AXIS_RIGHTY", "SDL_CONTROLLER_AXIS_TRIGGERLEFT",
    "SDL_CONTROLLER_AXIS_TRIGGERRIGHT", "SDL_CONTROLLER_AXIS_MAX",  

    "SDL_GameControllerButton",
    "SDL_CONTROLLER_BUTTON_INVALID", "SDL_CONTROLLER_BUTTON_A",
    "SDL_CONTROLLER_BUTTON_B", "SDL_CONTROLLER_BUTTON_X",
    "SDL_CONTROLLER_BUTTON_Y", "SDL_CONTROLLER_BUTTON_BACK",
    "SDL_CONTROLLER_BUTTON_GUIDE", "SDL_CONTROLLER_BUTTON_START",
    "SDL_CONTROLLER_BUTTON_LEFTSTICK", "SDL_CONTROLLER_BUTTON_RIGHTSTICK",
    "SDL_CONTROLLER_BUTTON_LEFTSHOULDER",
    "SDL_CONTROLLER_BUTTON_RIGHTSHOULDER",
    "SDL_CONTROLLER_BUTTON_DPAD_UP", "SDL_CONTROLLER_BUTTON_DPAD_DOWN",
    "SDL_CONTROLLER_BUTTON_DPAD_LEFT", "SDL_CONTROLLER_BUTTON_DPAD_RIGHT",
    "SDL_CONTROLLER_BUTTON_MISC1", "SDL_CONTROLLER_BUTTON_PADDLE1",
    "SDL_CONTROLLER_BUTTON_PADDLE2", "SDL_CONTROLLER_BUTTON_PADDLE3",
    "SDL_CONTROLLER_BUTTON_PADDLE4", "SDL_CONTROLLER_BUTTON_TOUCHPAD",
    "SDL_CONTROLLER_BUTTON_MAX",

    # Macro Functions
    "SDL_GameControllerAddMappingsFromFile",
    
    # Functions
    "SDL_GameControllerAddMappingsFromRW", "SDL_GameControllerAddMapping",
    "SDL_GameControllerNumMappings", "SDL_GameControllerMappingForIndex",
    "SDL_GameControllerMappingForGUID", "SDL_GameControllerMapping",
    "SDL_IsGameController", "SDL_GameControllerNameForIndex",
    "SDL_GameControllerTypeForIndex", "SDL_GameControllerMappingForDeviceIndex",
    "SDL_GameControllerOpen", "SDL_GameControllerFromInstanceID",
    "SDL_GameControllerFromPlayerIndex", "SDL_GameControllerName",
    "SDL_GameControllerGetType", 
    "SDL_GameControllerGetPlayerIndex", "SDL_GameControllerSetPlayerIndex",
    "SDL_GameControllerGetVendor", "SDL_GameControllerGetProduct",
    "SDL_GameControllerGetProductVersion", "SDL_GameControllerGetSerial",
    "SDL_GameControllerGetAttached", "SDL_GameControllerGetJoystick",
    "SDL_GameControllerEventState", "SDL_GameControllerUpdate",
    "SDL_GameControllerGetAxisFromString", "SDL_GameControllerGetStringForAxis",
    "SDL_GameControllerGetBindForAxis", "SDL_GameControllerHasAxis",
    "SDL_GameControllerGetAxis",
    "SDL_GameControllerGetButtonFromString",
    "SDL_GameControllerGetStringForButton",
    "SDL_GameControllerGetBindForButton", "SDL_GameControllerHasButton",
    "SDL_GameControllerGetButton",
    "SDL_GameControllerGetNumTouchpads",
    "SDL_GameControllerGetNumTouchpadFingers",
    "SDL_GameControllerGetTouchpadFinger",
    "SDL_GameControllerHasSensor", "SDL_GameControllerSetSensorEnabled",
    "SDL_GameControllerIsSensorEnabled", 
    "SDL_GameControllerGetSensorDataRate", "SDL_GameControllerGetSensorData",
    "SDL_GameControllerRumble", "SDL_GameControllerRumbleTriggers",
    "SDL_GameControllerHasLED", "SDL_GameControllerHasRumble",
    "SDL_GameControllerHasRumbleTriggers",
    "SDL_GameControllerSetLED",
    "SDL_GameControllerSendEffect",
    "SDL_GameControllerClose",
    "SDL_GameControllerGetAppleSFSymbolsNameForButton",
    "SDL_GameControllerGetAppleSFSymbolsNameForAxis",
]


class SDL_GameController(c_void_p):
    pass


SDL_GameControllerBindType = c_int

SDL_CONTROLLER_BINDTYPE_NONE = 0
SDL_CONTROLLER_BINDTYPE_BUTTON = 1
SDL_CONTROLLER_BINDTYPE_AXIS = 2
SDL_CONTROLLER_BINDTYPE_HAT = 3


SDL_GameControllerType = c_int

SDL_CONTROLLER_TYPE_UNKNOWN = 0
SDL_CONTROLLER_TYPE_XBOX360 = 1
SDL_CONTROLLER_TYPE_XBOXONE = 2
SDL_CONTROLLER_TYPE_PS3 = 3
SDL_CONTROLLER_TYPE_PS4 = 4
SDL_CONTROLLER_TYPE_NINTENDO_SWITCH_PRO = 5
SDL_CONTROLLER_TYPE_VIRTUAL = 6
SDL_CONTROLLER_TYPE_PS5 = 7
SDL_CONTROLLER_TYPE_AMAZON_LUNA = 8
SDL_CONTROLLER_TYPE_GOOGLE_STADIA = 9


class _gchat(Structure):
    _fields_ = [("hat", c_int), ("hat_mask", c_int)]

class _gcvalue(Union):
    _fields_ = [("button", c_int), ("axis", c_int), ("hat", _gchat)]

class SDL_GameControllerButtonBind(Structure):
    _fields_ = [("bindType", SDL_GameControllerBindType), ("value", _gcvalue)]


SDL_GameControllerAddMapping = _bind("SDL_GameControllerAddMapping", [c_char_p], c_int)
SDL_GameControllerMapping = _bind("SDL_GameControllerMapping", [POINTER(SDL_GameController)], c_char_p)
SDL_IsGameController = _bind("SDL_IsGameController", [c_int], SDL_bool)
SDL_GameControllerNameForIndex = _bind("SDL_GameControllerNameForIndex", [c_int], c_char_p)
SDL_GameControllerTypeForIndex = _bind("SDL_GameControllerTypeForIndex", [c_int], SDL_GameControllerType, added='2.0.12')
SDL_GameControllerOpen = _bind("SDL_GameControllerOpen", [c_int], POINTER(SDL_GameController))
SDL_GameControllerName = _bind("SDL_GameControllerName", [POINTER(SDL_GameController)], c_char_p)
SDL_GameControllerGetType = _bind("SDL_GameControllerGetType", [POINTER(SDL_GameController)], SDL_GameControllerType, added='2.0.12')
SDL_GameControllerGetAttached = _bind("SDL_GameControllerGetAttached", [POINTER(SDL_GameController)], SDL_bool)
SDL_GameControllerGetJoystick = _bind("SDL_GameControllerGetJoystick", [POINTER(SDL_GameController)], POINTER(SDL_Joystick))
SDL_GameControllerEventState = _bind("SDL_GameControllerEventState", [c_int], c_int)
SDL_GameControllerUpdate = _bind("SDL_GameControllerUpdate")


SDL_GameControllerAxis = c_int

SDL_CONTROLLER_AXIS_INVALID = -1
SDL_CONTROLLER_AXIS_LEFTX = 0
SDL_CONTROLLER_AXIS_LEFTY = 1
SDL_CONTROLLER_AXIS_RIGHTX = 2
SDL_CONTROLLER_AXIS_RIGHTY = 3
SDL_CONTROLLER_AXIS_TRIGGERLEFT = 4
SDL_CONTROLLER_AXIS_TRIGGERRIGHT = 5
SDL_CONTROLLER_AXIS_MAX = 6


SDL_GameControllerGetAxisFromString = _bind("SDL_GameControllerGetAxisFromString", [c_char_p], SDL_GameControllerAxis)
SDL_GameControllerGetStringForAxis = _bind("SDL_GameControllerGetStringForAxis", [SDL_GameControllerAxis], c_char_p)
SDL_GameControllerGetBindForAxis = _bind("SDL_GameControllerGetBindForAxis", [POINTER(SDL_GameController), SDL_GameControllerAxis], SDL_GameControllerButtonBind)
SDL_GameControllerHasAxis = _bind("SDL_GameControllerHasAxis", [POINTER(SDL_GameController), SDL_GameControllerAxis], SDL_bool, added='2.0.14')
SDL_GameControllerGetAxis = _bind("SDL_GameControllerGetAxis", [POINTER(SDL_GameController), SDL_GameControllerAxis], Sint16)


SDL_GameControllerButton = c_int

SDL_CONTROLLER_BUTTON_INVALID = -1
SDL_CONTROLLER_BUTTON_A = 0
SDL_CONTROLLER_BUTTON_B = 1
SDL_CONTROLLER_BUTTON_X = 2
SDL_CONTROLLER_BUTTON_Y = 3
SDL_CONTROLLER_BUTTON_BACK = 4
SDL_CONTROLLER_BUTTON_GUIDE = 5
SDL_CONTROLLER_BUTTON_START = 6
SDL_CONTROLLER_BUTTON_LEFTSTICK = 7
SDL_CONTROLLER_BUTTON_RIGHTSTICK = 8
SDL_CONTROLLER_BUTTON_LEFTSHOULDER = 9
SDL_CONTROLLER_BUTTON_RIGHTSHOULDER = 10
SDL_CONTROLLER_BUTTON_DPAD_UP = 11
SDL_CONTROLLER_BUTTON_DPAD_DOWN = 12
SDL_CONTROLLER_BUTTON_DPAD_LEFT = 13
SDL_CONTROLLER_BUTTON_DPAD_RIGHT = 14
SDL_CONTROLLER_BUTTON_MISC1 = 15    
SDL_CONTROLLER_BUTTON_PADDLE1 = 16
SDL_CONTROLLER_BUTTON_PADDLE2 = 17
SDL_CONTROLLER_BUTTON_PADDLE3 = 18
SDL_CONTROLLER_BUTTON_PADDLE4 = 19
SDL_CONTROLLER_BUTTON_TOUCHPAD = 20
SDL_CONTROLLER_BUTTON_MAX = 21
if sdl_version < 2014:
    SDL_CONTROLLER_BUTTON_MAX = 15  # For backwards compatibility


SDL_GameControllerGetButtonFromString = _bind("SDL_GameControllerGetButtonFromString", [c_char_p], SDL_GameControllerButton)
SDL_GameControllerGetStringForButton = _bind("SDL_GameControllerGetStringForButton", [SDL_GameControllerButton], c_char_p)
SDL_GameControllerGetBindForButton = _bind("SDL_GameControllerGetBindForButton", [POINTER(SDL_GameController), SDL_GameControllerButton], SDL_GameControllerButtonBind)
SDL_GameControllerHasButton = _bind("SDL_GameControllerHasButton", [POINTER(SDL_GameController), SDL_GameControllerButton], SDL_bool, added='2.0.14')
SDL_GameControllerGetButton = _bind("SDL_GameControllerGetButton", [POINTER(SDL_GameController), SDL_GameControllerButton], Uint8)
SDL_GameControllerGetNumTouchpads = _bind("SDL_GameControllerGetNumTouchpads", [POINTER(SDL_GameController)], c_int, added='2.0.14')
SDL_GameControllerGetNumTouchpadFingers = _bind("SDL_GameControllerGetNumTouchpadFingers", [POINTER(SDL_GameController), c_int], c_int, added='2.0.14')
SDL_GameControllerGetTouchpadFinger = _bind("SDL_GameControllerGetTouchpadFinger", [POINTER(SDL_GameController), c_int, c_int, POINTER(Uint8), POINTER(c_float), POINTER(c_float), POINTER(c_float)], c_int, added='2.0.14')
SDL_GameControllerHasSensor = _bind("SDL_GameControllerHasSensor", [POINTER(SDL_GameController), SDL_SensorType], SDL_bool, added='2.0.14')
SDL_GameControllerSetSensorEnabled = _bind("SDL_GameControllerSetSensorEnabled", [POINTER(SDL_GameController), SDL_SensorType, SDL_bool], c_int, added='2.0.14')
SDL_GameControllerIsSensorEnabled = _bind("SDL_GameControllerIsSensorEnabled", [POINTER(SDL_GameController), SDL_SensorType], SDL_bool, added='2.0.14')
SDL_GameControllerGetSensorDataRate = _bind("SDL_GameControllerGetSensorDataRate", [POINTER(SDL_GameController), SDL_SensorType], c_float, added='2.0.16')
# TODO: Read how GetSensorData is implemented to figure out how the # of floats is determined
SDL_GameControllerGetSensorData = _bind("SDL_GameControllerGetSensorData", [POINTER(SDL_GameController), SDL_SensorType, POINTER(c_float), c_int], c_int, added='2.0.14')
SDL_GameControllerAddMappingsFromRW = _bind("SDL_GameControllerAddMappingsFromRW", [POINTER(SDL_RWops), c_int], c_int)
SDL_GameControllerAddMappingsFromFile = lambda fname: SDL_GameControllerAddMappingsFromRW(SDL_RWFromFile(fname, b"rb"), 1)
SDL_GameControllerFromInstanceID = _bind("SDL_GameControllerFromInstanceID", [SDL_JoystickID], POINTER(SDL_GameController), added='2.0.4')
SDL_GameControllerFromPlayerIndex = _bind("SDL_GameControllerFromPlayerIndex", [c_int], POINTER(SDL_GameController), added='2.0.12')
SDL_GameControllerGetPlayerIndex = _bind("SDL_GameControllerGetPlayerIndex", [POINTER(SDL_GameController)], c_int, added='2.0.9')
SDL_GameControllerSetPlayerIndex = _bind("SDL_GameControllerSetPlayerIndex", [POINTER(SDL_GameController), c_int], added='2.0.12')
SDL_GameControllerGetVendor = _bind("SDL_GameControllerGetVendor", [POINTER(SDL_GameController)], Uint16, added='2.0.6')
SDL_GameControllerGetProduct = _bind("SDL_GameControllerGetProduct", [POINTER(SDL_GameController)], Uint16, added='2.0.6')
SDL_GameControllerGetProductVersion = _bind("SDL_GameControllerGetProductVersion", [POINTER(SDL_GameController)], Uint16, added='2.0.6')
SDL_GameControllerGetSerial = _bind("SDL_GameControllerGetSerial", [POINTER(SDL_GameController)], c_char_p, added='2.0.14')
SDL_GameControllerNumMappings = _bind("SDL_GameControllerNumMappings", None, c_int, added='2.0.6')
SDL_GameControllerMappingForIndex = _bind("SDL_GameControllerMappingForIndex", [c_int], c_char_p, added='2.0.6')
SDL_GameControllerMappingForDeviceIndex = _bind("SDL_GameControllerMappingForDeviceIndex", [c_int], c_char_p, added='2.0.9')
SDL_GameControllerRumble = _bind("SDL_GameControllerRumble", [POINTER(SDL_GameController), Uint16, Uint16, Uint32], c_int, added='2.0.9')
SDL_GameControllerRumbleTriggers = _bind("SDL_GameControllerRumbleTriggers", [POINTER(SDL_GameController), Uint16, Uint16, Uint32], c_int, added='2.0.14')
SDL_GameControllerHasLED = _bind("SDL_GameControllerHasLED", [POINTER(SDL_GameController)], SDL_bool, added='2.0.14')
SDL_GameControllerHasRumble = _bind("SDL_GameControllerHasRumble", [POINTER(SDL_GameController)], SDL_bool, added='2.0.18')
SDL_GameControllerHasRumbleTriggers = _bind("SDL_GameControllerHasRumbleTriggers", [POINTER(SDL_GameController)], SDL_bool, added='2.0.18')
SDL_GameControllerSetLED = _bind("SDL_GameControllerSetLED", [POINTER(SDL_GameController), Uint8, Uint8, Uint8], c_int, added='2.0.14')
SDL_GameControllerSendEffect = _bind("SDL_GameControllerSendEffect", [POINTER(SDL_GameController), c_void_p, c_int], c_int, added='2.0.16')
SDL_GameControllerClose = _bind("SDL_GameControllerClose", [POINTER(SDL_GameController)])
SDL_GameControllerGetAppleSFSymbolsNameForButton = _bind("SDL_GameControllerGetAppleSFSymbolsNameForButton", [POINTER(SDL_GameController), SDL_GameControllerButton], c_char_p, added='2.0.18')
SDL_GameControllerGetAppleSFSymbolsNameForAxis = _bind("SDL_GameControllerGetAppleSFSymbolsNameForAxis", [POINTER(SDL_GameController), SDL_GameControllerAxis], c_char_p, added='2.0.18')

# Reimplemented w/ other functions due to crash-causing ctypes bug (fixed in 3.8)
if sys.version_info >= (3, 8, 0, 'final'):
    SDL_GameControllerMappingForGUID = _bind("SDL_GameControllerMappingForGUID", [SDL_JoystickGUID], c_char_p)
else:
    def SDL_GameControllerMappingForGUID(guid):
        buff = create_string_buffer(33)
        SDL_JoystickGetGUIDString(guid, buff, 33) # Get GUID string
        guid_str = buff.value
        # Iterate over controller mappings and look for a GUID match
        # Note: iterates in reverse, so user-defined mappings are checked first
        num = SDL_GameControllerNumMappings()
        for i in range(num - 1, -1, -1): 
            m = SDL_GameControllerMappingForIndex(i)
            if m.split(b',')[0] == guid_str:
                return m
        return None



import sys
import pytest
from sdl2 import ext as sdl2ext
from sdl2 import SDL_WasInit, SDL_INIT_VIDEO, SDL_FlushEvent, SDL_USEREVENT, \
    SDL_FIRSTEVENT, SDL_LASTEVENT, SDL_Event, SDL_UserEvent, SDL_PushEvent

@pytest.fixture(scope="module")
def with_sdl_ext():
    sdl2ext.init()
    yield
    sdl2ext.quit()


def test_init_quit():
    # NOTE: Currently init only inits the video subsystem, but quit shuts down
    # SDL2 and ttf/image/mixer libraries. This latter function should be tested.
    try:
        sdl2ext.init()
    except sdl2ext.SDLError:
        raise pytest.skip('Video subsystem not supported')
    assert SDL_WasInit(SDL_INIT_VIDEO) == SDL_INIT_VIDEO
    sdl2ext.quit()
    assert SDL_WasInit(SDL_INIT_VIDEO) != SDL_INIT_VIDEO
    sdl2ext.init()
    sdl2ext.init()
    sdl2ext.init()
    assert SDL_WasInit(SDL_INIT_VIDEO) == SDL_INIT_VIDEO
    sdl2ext.quit()
    assert SDL_WasInit(SDL_INIT_VIDEO) != SDL_INIT_VIDEO

def test_get_events(with_sdl_ext):
    SDL_FlushEvent(SDL_FIRSTEVENT, SDL_LASTEVENT)
    for x in range(12):
        event = SDL_Event()
        event.type = SDL_USEREVENT + x
        event.user = SDL_UserEvent(
            type=event.type, timestamp=0, windowID=0, code=0
        )
        SDL_PushEvent(event)
    results = sdl2ext.get_events()
    assert len(results) == 12
    for idx, r in enumerate(results):
        assert idx == r.type - SDL_USEREVENT

def test_TestEventProcessor(with_sdl_ext):
    # NOTE: This doesn't really test functionality, but since I don't think
    # it's terribly useful I'm not going to bother expanding it
    proc = sdl2ext.TestEventProcessor()
    assert isinstance(proc, sdl2ext.TestEventProcessor)



import ctypes
from .. import SDL_Init, SDL_Quit, SDL_QuitSubSystem, SDL_WasInit, \
    SDL_INIT_VIDEO, error, events, timer

_HASSDLTTF = True
try:
    from .. import sdlttf
except ImportError:
    _HASSDLTTF = False
_HASSDLIMAGE = True
try:
    from .. import sdlimage
except ImportError:
    _HASSDLIMAGE = False

__all__ = ["SDLError", "init", "quit", "get_events", "TestEventProcessor"]


class SDLError(Exception):
    """An SDL2-specific exception class."""

    def __init__(self, msg=None):
        """Creates a new SDLError instance with the specified message.

        If no msg is passed, it will try to get the current SDL2 error via
        :func:`sdl2.SDL_GetError`.

        """
        super(SDLError, self).__init__()
        self.msg = msg
        if not msg:
            self.msg = error.SDL_GetError()
            error.SDL_ClearError()

    def __str__(self):
        return repr(self.msg)


def raise_sdl_err(desc=None):
    # Raises and clears the latest SDL error. For internal use.
    errmsg = error.SDL_GetError().decode('utf-8')
    error.SDL_ClearError()
    e = "Error encountered"
    if desc:
        e += " " + desc
    if len(errmsg):
        e += ": {0}".format(errmsg)
    raise SDLError(e)


def init():
    """Initializes the SDL2 video subsystem.

    See :ref:`pygamers_pygame` for a comparison between this function and
    ``pygame.init()``.

    Raises:
        :exc:`SDLError`: If the SDL2 video subsystem cannot be initialized.

    """
    # TODO: More subsystems?
    if SDL_Init(SDL_INIT_VIDEO) != 0:
        raise SDLError()


def quit():
    """Quits the SDL2 video subysystem.

    If no other subsystems are active, this will also call
    :func:`sdl2.SDL_Quit`, :func:`sdlttf.TTF_Quit` and
    :func:`sdlimage.IMG_Quit`.

    """
    # TODO: More subsystems? Also, is TTF_WasInit always 1?
    SDL_QuitSubSystem(SDL_INIT_VIDEO)
    if SDL_WasInit(0) != 0:
        if _HASSDLTTF and sdlttf.TTF_WasInit() > 0:
            sdlttf.TTF_Quit()
        if _HASSDLIMAGE:
            sdlimage.IMG_Quit()
        SDL_Quit()


def get_events():
    """Gets all SDL events that are currently on the event queue.

    Returns:
        :obj:`List`: A list of all :obj:`~sdl2.SDL_Event` objects currently in
        the event queue.
    
    """
    events.SDL_PumpEvents()

    evlist = []
    SDL_PeepEvents = events.SDL_PeepEvents

    op = events.SDL_GETEVENT
    first = events.SDL_FIRSTEVENT
    last = events.SDL_LASTEVENT

    while True:
        evarray = (events.SDL_Event * 10)()
        ptr = ctypes.cast(evarray, ctypes.POINTER(events.SDL_Event))
        ret = SDL_PeepEvents(ptr, 10, op, first, last)
        if ret <= 0:
            break
        evlist += list(evarray)[:ret]
        if ret < 10:
            break

    return evlist


class TestEventProcessor(object):
    """A simple event processor for testing purposes."""

    def run(self, window):
        """Starts an event loop without actually processing any event.
        
        This method will run endlessly until an ``SDL_QUIT`` event occurs.

        Args:
            window (:obj:`sdl2.ext.Window`): The window within which to run
                the test event loop.
        
        """
        event = events.SDL_Event()
        running = True
        while running:
            ret = events.SDL_PollEvent(ctypes.byref(event), 1)
            if ret == 1:
                if event.type == events.SDL_QUIT:
                    running = False
                    break
            window.refresh()
            timer.SDL_Delay(10)



from ctypes import byref, c_int
from .color import Color
from .compat import isiterable, utf8
from .common import raise_sdl_err
from .window import Window
from .. import dll, SDL_PumpEvents, SDL_Window
from .. import messagebox as mb

__all__ = [
    "MessageBoxTheme", "MessageBox", "show_messagebox", "show_alert"
]

class MessageBoxTheme(object):
    """Initializes a color scheme for use with :obj:`MessageBox` objects.

    This is used to define the background, text, and various button colors
    to use when presenting dialog boxes to users. All colors must be defined
    as either :obj:`sdl2.ext.Color` objects or 8-bit ``(r, g, b)`` tuples.

    .. note::
       SDL2 only supports MessageBox themes on a few platforms, including
       Linux/BSD (if using X11) and Haiku. MessageBox themes will have no effect
       on Windows, macOS, or Linux if using Wayland.

    Args:
        bg (:obj:`~sdl2.ext.Color`, tuple, optional): The color to use for the
            background of the dialog box. Defaults to ``(56, 54, 53)``.
        text (:obj:`~sdl2.ext.Color`, tuple, optional): The color to use for the
            text of the dialog box. Defaults to ``(209, 207, 205)``.
        btn (:obj:`~sdl2.ext.Color`, tuple, optional): The color to use for the
            backgrounds of buttons. Defaults to ``(140, 135, 129)``.
        btn_border (:obj:`~sdl2.ext.Color`, tuple, optional): The color to use
            for the borders of buttons. Defaults to ``(105, 102, 99)``.
        btn_selected (:obj:`~sdl2.ext.Color`, tuple, optional): The color to use
            for selected buttons. Defaults to ``(205, 202, 53)``.

    """
    def __init__(
        self, bg=None, text=None, btn=None, btn_border=None, btn_selected=None
    ):
        # NOTE: Default colors taken from SDL_x11messagebox.c
        self._theme = [
            (56, 54, 53),     # Background color
            (209, 207, 205),  # Text color
            (140, 135, 129),  # Button border color
            (105, 102, 99),   # Button background color
            (205, 202, 53)    # Selected button color
        ]
        # Update default theme colors based on provided values
        elements = [bg, text, btn_border, btn, btn_selected]
        for i in range(len(elements)):
            if elements[i] is not None:
                self._theme[i] = self._validate_color(elements[i])

    def _validate_color(self, col):
        if not isinstance(col, Color):
            if not isiterable(col) or len(col) != 3:
                e = "MessageBox colors must be specified as (r, g, b) tuples."
                raise TypeError(e)
            for val in col:
                if int(val) != float(val):
                    e = "All RGB values must be integers between 0 and 255."
                    raise ValueError(e)
            col = Color(col[0], col[1], col[2])
        return (col.r, col.g, col.b)

    def _get_theme(self):
        sdl_colors = []
        for col in self._theme:
            sdl_colors.append(mb.SDL_MessageBoxColor(*col))
        col_array = (mb.SDL_MessageBoxColor * 5)(*sdl_colors)
        return mb.SDL_MessageBoxColorScheme(col_array)


class MessageBox(object):
    """Creates a prototype for a dialog box that can be presented to the user.

    The `MessageBox` class is for designing a dialog box in the style of the
    system's window manager, containing a title, a message to present, and
    one or more response buttons.

    Args:
        title (str): The title to use for the dialog box. All UTF-8 characters
            are supported.
        msg (str): The main body of text to display in the dialog box. All UTF-8
            characters are supported.
        buttons (list): A list of strings, containing the labels of the buttons
            to place at the bottom of the dialog box (e.g. ``["No", "Yes"]``).
            Buttons will be placed in left-to-right order.
        default (str, optional): The label of the button to highlight as the
            default option (e.g. ``"Yes"``). Must match one of the labels in
            ``buttons``. This option will be accepted if the Return/Enter key
            is pressed on the keyboard.
        msgtype (str, optional): The type of dialog box to create, if supported
            by the system. On most window managers, this changes the icon used
            in the dialog box. Must be one of 'error', 'warning', or 'info', or
            None (the default).
        theme (:obj:`MessageBoxTheme`, optional): The color scheme to use for
            the dialog box, if supported by the window manager. Defaults to the
            system default theme.

    """
    def __init__(self, title, msg, buttons, default=None, msgtype=None, theme=None):
        self._title = utf8(title).encode('utf-8')
        self._text = utf8(msg).encode('utf-8')
        self._validate_buttons(buttons)
        self._buttons = buttons
        self._sdlbuttons = self._init_buttons(buttons, default)
        self._type = self._set_msgtype(msgtype) if msgtype else 0
        self._theme = theme._get_theme() if theme else None

    def _set_msgtype(self, msgtype):
        _flagmap = {
            'error': mb.SDL_MESSAGEBOX_ERROR,
            'warning': mb.SDL_MESSAGEBOX_WARNING,
            'info': mb.SDL_MESSAGEBOX_INFORMATION,
        }
        if msgtype.lower() not in _flagmap.keys():
            raise ValueError(
                "MessageBox type must be 'error', 'warning', 'info', or None."
            )
        return _flagmap[msgtype]

    def _validate_buttons(self, buttons):
        if not isiterable(buttons):
            raise TypeError("Buttons must be provided as a list.")
        elif len(buttons) == 0:
            raise ValueError("MessageBox must have at least one button.")

    def _init_buttons(self, buttons, default):
        default_flag = mb.SDL_MESSAGEBOX_BUTTON_RETURNKEY_DEFAULT
        buttonset = []
        for i in range(len(buttons)):
            b = mb.SDL_MessageBoxButtonData(
                flags = (default_flag if buttons[i] == default else 0),
                buttonid = i,
                text = utf8(buttons[i]).encode('utf-8'),
            )
            buttonset.append(b)
        return (mb.SDL_MessageBoxButtonData * len(buttons))(*buttonset)

    def _get_window_pointer(self, win):
        if isinstance(win, Window):
            win = win.window
        if isinstance(win, SDL_Window):
            win = dll.get_pointer(win)
        if hasattr(win, "contents") and isinstance(win.contents, SDL_Window):
            return win
        else:
            e = "'window' must be a Window or SDL_Window object (got {0})"
            raise ValueError(e.format(str(type(win))))

    def _get_msgbox(self, window=None):
        if window:
            window = self._get_window_pointer(window)
        return mb.SDL_MessageBoxData(
            flags = self._type | mb.SDL_MESSAGEBOX_BUTTONS_RIGHT_TO_LEFT,
            window = window,
            title = self._title,
            message = self._text,
            numbuttons = len(self._buttons),
            buttons = self._sdlbuttons,
            colorScheme = dll.get_pointer(self._theme) if self._theme else None,
        )
        

def show_messagebox(msgbox, window=None):
    """Displays a dialog box to the user and waits for a response.

    By default message boxes are presented independently of any window, but
    they can optionally be attached explicitly to a specific SDL window. This
    prevents that window from regaining focus until a response to the dialog
    box is made.

    Args:
        msgbox (:obj:`~sdl2.ext.MessageBox`): The dialog box to display
            on-screen.
        window (:obj:`~sdl2.SDL_Window`, :obj:`~sdl2.ext.Window`, optional): The
            window to associate with the dialog box. Defaults to None.

    Returns:
        str: The label of the button selected by the user.

    """
    resp = c_int(-1)
    ret = mb.SDL_ShowMessageBox(
        msgbox._get_msgbox(window),
        byref(resp)
    )
    SDL_PumpEvents()
    if ret == 0:
        return msgbox._buttons[resp.value]
    else:
        raise_sdl_err("displaying the message box")


def show_alert(title, msg, msgtype=None, window=None):
    """Displays a simple alert to the user and waits for a response.

    This function is a simplified version of :func:`show_messagebox` for cases
    where only one response button ("OK") is needed and a custom color scheme
    is not necessary.

    By default message boxes are presented independently of any window, but
    they can optionally be attached explicitly to a specific SDL window. This
    prevents that window from regaining focus until a response to the dialog
    box is made.

    Args:
        msgbox (:obj:`~sdl2.ext.MessageBox`): The dialog box to display
            on-screen.
        window (:obj:`~sdl2.SDL_Window`, :obj:`~sdl2.ext.Window`, optional): The
            window to associate with the dialog box. Defaults to ``None``.

    """
    box = MessageBox(title, msg, ["OK"], msgtype=msgtype)
    if window:
        window = box._get_window_pointer(window)
    ret = mb.SDL_ShowSimpleMessageBox(
        box._type,
        box._title,
        box._text,
        window
    )
    SDL_PumpEvents()
    if ret < 0:
        raise_sdl_err("displaying the message box")
